/*
  genholon - evaluate 4x4 matrix for holonomy, given x,y,z on S^2, theta_hol

   Copyright (C) 2005 Boud Roukema
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
     See also http://www.gnu.org/licenses/gpl.html
*/

#include <stdlib.h>
#include "astromisc.h"
#include "genholon.h"



/* which layer is PDS holonomy i in ? */
int layer(int layer_index[MAX_LAYER][2],  /* INPUT, do not modify */
	  int i /* input */
	  ){
  int i_l;

  i_l = 0; /* we need to start with the identity */
  /*  printf("layer: i = %d i_l = %d \n",i,i_l); */
  while ( i_l < 10-1 && 
	  (i < layer_index[i_l][0] || layer_index[i_l][1] < i)){
    i_l += 1;
    /*    printf("layer: i = %d i_l = %d \n",i,i_l); */
  };
  /*  printf("layer: done\n"); */
  return i_l ;
};


/* calculate the pairwise products of holonomies as 4x4 matrices and
   find out which element of the group results */
int product_table( int i_space,                    /* which space ? */
		   double holonomies_R4[MAX_ORDER][4][4], 
		   int holonomy_products_i[MAX_ORDER][MAX_ORDER] ){

  double product_R4[4][4]; /* product as a 4x4 matrix */
  const double tol_test2_sq = 1e-10 ;
  int i,j; /* multiply holonomy i (left) by holonomy j (right) */
  int k;   /* find the holonomy k corresponding to the product */

  int ii,jj; /* iterators inside a matrix */

  double xx,dd; /* temporary work */

  int test = 1;

  for (i=0; i< order[i_space]; i++){
    for (j=0; j< order[i_space]; j++){

      /* multiply holonomy i (left) by holonomy j (right) */
      cblas_dgemm(101, /* CblasRowMajor */
		  111, 111, /* no transpose */
		  4,4,4, 1.0, 
		  (double *)&(holonomies_R4[i][0][0]), 4, 
		  (double *)&(holonomies_R4[j][0][0]), 4,
		  0.0, 
		  (double *)&product_R4, 4);  

      /* find out which holonomy the product is close to, otherwise
	 flag error */
      xx = 9e9;
      k = 0;
      while (k < order[i_space] && xx > tol_test2_sq ){
	xx = 0.0;
	for (ii=0; ii< 4; ii++){
	  for (jj=0; jj< 4; jj++){
	    dd = product_R4[ii][jj] - holonomies_R4[k][ii][jj] ;
	    xx += dd*dd;
	  };
	};
	xx /= 16.0;
	/*	if(test){
	  printf(" i = %d  j= %d  xx = %g  k= %d\n",i,j,xx,k);
	  }; 
	*/
	k ++;
      };
      if (xx > tol_test2_sq){
	printf("ERROR: product_table: product not found i = %d  j= %d  k= %d\n",
	       i,j,k);
      }
      else{
	holonomy_products_i[i][j] = k-1;
      };
    };  /*  for (j=0; j< order[i_space]; j++) */

    if(test){
      /* print a subsample of the first few products for each i to stdout */
      printf("%3d: ",i);
      for (j=0; j< 20; j++){  
	printf(" %3d",holonomy_products_i[i][j]);
      };
      printf("\n");
    };

  };  /*   for (i=0; i< order[i_space]; i++) */
  return (0);
};

/* calculate the holonomies as 4x4 matrices generated by a fixed
   pair of would-be generators, find out which element of the group 
   results in each case, in order to see if the full group is
   generated */
int product_table_generator_pair(int i_space,                    /* which space ? */ 
                                 double holonomies_R4[MAX_ORDER][4][4], 
                                 int generators[2],
                                 int max_gen_order
                                 ){

  double product_R4[4][4]; /* product as a 4x4 matrix */

  double holonomy_products_R4[max_gen_order][max_gen_order][4][4]; 
  /* 
     ** if too big to declare directly, then more complicated 
     to handle big multi-dimensional array, e.g. allocate with malloc **
     double *holonomy_products_R4;
  */
  int holonomy_products_i[max_gen_order][max_gen_order];

  const double tol_test2_sq = 1e-10 ;
  int i,j; /* multiply generators[0]^i (left) by generators[1]^j (right) */
  int k;   /* find the holonomy k corresponding to the product */

  int ii,jj; /* iterators inside a matrix */

  double xx,dd; /* temporary work */

  int test = 1;

  /*  holonomy_products_R4 = malloc(MAX_ORDER*MAX_ORDER*4*4*sizeof(double)); */

  /* powers of generators[0] */
  for (i=1; i< max_gen_order; i++){
    if(1==i){
      for (ii=0; ii< 4; ii++){
        for (jj=0; jj< 4; jj++){
          holonomy_products_R4[0][i][ii][jj] = holonomies_R4[generators[0]][ii][jj] ;
        };
      };
    }else{
      /* multiply by the generator to get next power */
      cblas_dgemm(101, /* CblasRowMajor */
                  111, 111, /* no transpose */
                  4,4,4, 1.0, 
                  (double *)&(holonomy_products_R4[0][i-1][0][0]), 4, 
                  (double *)&(holonomy_products_R4[0][1][0][0]), 4, 
                  0.0, 
                  (double *)&(holonomy_products_R4[0][i][0][0]), 4);  
    };
  };

  /* powers of generators[1] */
  for (j=1; j< max_gen_order; j++){
    if(1==j){
      for (ii=0; ii< 4; ii++){
        for (jj=0; jj< 4; jj++){
          holonomy_products_R4[j][0][ii][jj] = holonomies_R4[generators[1]][ii][jj] ;
        };
      };
    }else{
      /* multiply by the generator to get next power */
      cblas_dgemm(101, /* CblasRowMajor */
                  111, 111, /* no transpose */
                  4,4,4, 1.0, 
                  (double *)&(holonomy_products_R4[j-1][0][0][0]), 4, 
                  (double *)&(holonomy_products_R4[1][0][0][0]), 4, 
                  0.0, 
                  (double *)&(holonomy_products_R4[j][0][0][0]), 4);  
    };
  };


  /* assume order of group is correct => order[]-1 */
  for (i=0; i< max_gen_order-1; i++){  
    for (j=0; j< max_gen_order-1; j++){

      if(i>0 && 0==j){
        for (ii=0; ii< 4; ii++){
          for (jj=0; jj< 4; jj++){
            product_R4[ii][jj] = holonomy_products_R4[0][i][ii][jj];
          };
        };
      }else if(0==i && j>0){
        for (ii=0; ii< 4; ii++){
          for (jj=0; jj< 4; jj++){
            product_R4[ii][jj] = holonomy_products_R4[j][0][ii][jj];
          };
        };
      }else if(i>0 && j>0){

        /* multiply generators[0]^i (left) by generators[1]^j (right) */
        cblas_dgemm(101, /* CblasRowMajor */
                    111, 111, /* no transpose */
                    4,4,4, 1.0, 
                    (double *)&(holonomy_products_R4[0][i][0][0]), 4, 
                    (double *)&(holonomy_products_R4[j][0][0][0]), 4, 
                    0.0, 
                    (double *)&product_R4, 4);  
      };

      if(i>0 || j>0){ /* no point checking identity */
        /* find out which holonomy the product is close to, otherwise
           flag error */
        xx = 9e9;
        k = 0;
        while (k < order[i_space] && xx > tol_test2_sq ){
          xx = 0.0;
          for (ii=0; ii< 4; ii++){
            for (jj=0; jj< 4; jj++){
              dd = product_R4[ii][jj] - holonomies_R4[k][ii][jj] ;
              xx += dd*dd;
            };
          };
          xx /= 16.0;
          /*	if(test){
                printf(" i = %d  j= %d  xx = %g  k= %d\n",i,j,xx,k);
                }; 
          */
          k ++;
        };
        if (xx > tol_test2_sq){
          printf("ERROR: product_table_generator_pair: product not found i = %d  j= %d  k= %d\n",
                 i,j,k);
        }
        else{
          holonomy_products_i[i][j] = k-1;
        };
      }else{ /* else if(i>0 && j>0) */
        holonomy_products_i[0][0] = 0;  /* TOCHECK: assumes identity is 0th holonomy */
      };
    };  /*  for (j=0; j< max_gen_order-1; j++) */

    if(test){
      /* print a subsample of the first few products for each i to stdout */
      printf("gen: ");
      printf("%3d: ",i);
      /*      for (j=0; j< 20; j++){   */
      for (j=0; j< max_gen_order-1; j++){  
	printf(" %3d",holonomy_products_i[i][j]);
      };
      printf("\n");
    };

  };  /*   for (i=0; i< max_gen_order-1; i++) */

  /* free(holonomy_products_R4); */
  return (0);
};




/* check holonomy generating functions */
void CHKHOL(void){

  double holonomies_R4[MAX_ORDER][4][4]; /* holonomies as 4x4 matrices */
  int holonomy_products_i[MAX_ORDER][MAX_ORDER]; /* pairwise products of holonomies */

#define N_ARBTWIST 7  
  double arb_twist[N_ARBTWIST] = {0.2*M_PI, -0.2*M_PI, 1e-5, 1e-4, 1e-3, 1e-2, 0.1}; /* arbitrary twist */

  int test= 0;
  int i_space;
  int i;
#define GENERATOR_PAIR {1,57}
  int generators[2] = GENERATOR_PAIR;
  
  for (i_space=0; i_space< N_SPACES; i_space++){
    if (test){
      printf("\ni_space = %d, space = %s\n",
	     i_space, spaces[i_space]);      
	     /*	     i_space, space_name(i_space));    TEST ONLY */
    };
    /* this generates hols, but also tests roots of identity */
    genholon(i_space, holonomies_R4, 
             184.0, 62.0, 34.0, 
             -1, 0, 77.77);   /* 77.77 should be ignored */

    if(test){
      printf("Will call check product_table:... \n");
      product_table(i_space, holonomies_R4, holonomy_products_i);
      printf(" ...product_table done, including error reports of");
      printf(" products outside of the group.\n");

      printf("Will call check product_table_generator_pair:... \n");
      product_table_generator_pair(i_space, holonomies_R4, generators, 11);
      printf(" ...product_table done, including error reports of");
      printf(" products outside of the group.\n");
    };
  };
  
  i_space= 2; /* S^3/I^* */

  /* test twist angles of -M_PI/5 and epsilon << 1 */
  
  for(i=0; i<N_ARBTWIST; i++){
    printf("checking isometry_R3_R4_twist generator...  arb_twist = %.5f\n",
           arb_twist[i]);
    genholon(i_space, holonomies_R4, 
             184.0, 62.0, 34.0, 
             1, 1, arb_twist[i]); 
    printf("\n");
  };

};


/* calculate the 120 holonomies as 4x4 matrices */
int genholon( int i_space,                             /* which space */ 
	      double holonomies_R4[MAX_ORDER][4][4],
              double gal_l, /* face centre if dodecahedral space */
              double gal_b,
              double theta, /* third orientation parameter if dodecspace */
              int    handedness,  /* -1 = lefthanded, +1 = righthanded */
              int    allow_arb_twist, /* 0 = don't allow, 1 = allow */
              double arb_twist  /* arbitrary twist angle */
              ){ 

  /* S^3/Gamma symmetries, layers */
  double symm[MAX_ORDER/2][3];
  int    layer_index[MAX_LAYER][2];
  int    ntwist[MAX_LAYER/2];

  /*  float gall = 137.0 ; */ /* arbitrary longitude */
  /* float galb = -59.0 ; */ /* arbitrary latitude */
  /* float theta_dodec = 17.0 ; */ /* arbitrary 3rd angle */
  /* const float pi_on_5 = M_PI / 5.0;  */ /* radians */
  /* const float pi_on_5_deg = 36.0 ;   */ /* degrees */
  const double TOL_SQRT = 1e-20; 
  /* float xdod[12];
  float ydod[12];
  float zdod[12];
  float xdodp[2][12];
  float ydodp[2][12];
  float zdodp[2][12]; */

  double points_R3_SLS[MAX_ORDER][3];  /* x,y,z where R_SLS = 1 */
  /*  double holonomies_R4[MAX_ORDER][4][4]; */ /* holonomies as 4x4 matrices */
  double points_tmp[3]; /* temporary copy of rotated points */

  double a_norm; /* a norm */
  /*  double ang;  */

  int ierr; /* error code */
  int i,ihol;
  int j,i_l,i_l_neg; /* for finding the layer */
  int test= 0;

  /*  GENDODEC(&gall,&galb,&theta_dodec,
	   xdod,ydod,zdod,
	   xdodp,ydodp,zdodp);
  */


  /* copy chosen holonomy (TODO: in principle this could be done
     faster, more abstractly, without copying, i.e. with indexing,
     struct, whatever; this might require filling initialisation
     arrays with lots of zeroes) */
  for (ihol=0; ihol < order[i_space]/2; ihol ++){
    for (j=0; j<3; j++){
      if (i_space == 0){
	symm[ihol][j] = octa_symm[ihol][j];
      }else if (i_space == 1){
	symm[ihol][j] = tcube_symm[ihol][j];
      }else if (i_space == 2){
	symm[ihol][j] = dodec_symm[ihol][j];
      }else{
	printf("Space number %d not implemented.\n",i_space);
      };
    };
  };
  for (i_l=0; i_l < n_layers[i_space]; i_l ++){
    for (j=0; j<2; j++){
      if (i_space == 0){
	layer_index[i_l][j] = octa_layer_index[i_l][j];
      }else if (i_space == 1){
	layer_index[i_l][j] = tcube_layer_index[i_l][j];
      }else if (i_space == 2){
	layer_index[i_l][j] = dodec_layer_index[i_l][j];
      }else{
	printf("Space number %d not implemented.\n",i_space);
      };

    };
  };
  for (i_l=0; i_l < n_layers[i_space]/2; i_l ++){

    if (i_space == 0){
      ntwist[i_l] = octa_ntwist[i_l];
    }else if (i_space == 1){
      ntwist[i_l] = tcube_ntwist[i_l];
    }else if (i_space == 2){
      ntwist[i_l] = dodec_ntwist[i_l];
    }else{
      printf("Space number %d not implemented.\n",i_space);
    };
  };




  /*  for (i=0; i<12; i++)     */
  for (i=1; i< order[i_space]/2; i++){   
    /* points_R3_SLS[i][0] = xdod[i];
    points_R3_SLS[i][1] = ydod[i];
    points_R3_SLS[i][2] = zdod[i]; */

    /*      a_norm = sqrt(PHI_GOLDEN*PHI_GOLDEN + 1.0); */

    a_norm = symm[i][0] * symm[i][0] + 
      symm[i][1] * symm[i][1] + 
      symm[i][2] * symm[i][2] ;
    a_norm = (a_norm > TOL_SQRT) ? sqrt(a_norm) : sqrt(TOL_SQRT) ;

    points_R3_SLS[i][0] = symm[i][0] / a_norm ;
    points_R3_SLS[i][1] = symm[i][1] / a_norm ;
    points_R3_SLS[i][2] = symm[i][2] / a_norm ;

    if (i_space == 2){  /* orient in astronomical coords; astromisc-0.1.18.2 */
      rotate_dodec_new2old(
                           gal_l,gal_b,theta,
                           points_R3_SLS[i][0],
                           points_R3_SLS[i][1],
                           points_R3_SLS[i][2],
                           &(points_tmp[0]),
                           &(points_tmp[1]),
                           &(points_tmp[2]) );
      points_R3_SLS[i][0] = points_tmp[0];
      points_R3_SLS[i][1] = points_tmp[1];
      points_R3_SLS[i][2] = points_tmp[2];
    };


    if(test){
      printf("\ni_holonomy= %d\n",i);
      printf("   using %f  %f  %f \n",symm[i][0],
	     symm[i][1],symm[i][2]);
    };

    i_l = layer(layer_index, i); /* find which layer this holonomy is in */
    if(test){
      printf("i = %d  i_l = %d \n",i,i_l);
      printf("ntwist[i_l] = %d \n",ntwist[i_l]);
    };

    if(!allow_arb_twist){
      ierr = holonomy_R3_R4( (void *)&(points_R3_SLS[i][0]), 
                             /* pi_on_5 if PDS 1st layer*/
                             (2.0* M_PI /(double)ntwist[i_l]),
                             handedness,
                             (void *)&(holonomies_R4[i][0][0])
                             ); 
    }else{
      ierr = isometry_R3_R4_twist( (void *)&(points_R3_SLS[i][0]), 
                                   /* pi_on_5 if PDS 1st layer*/
                                   (2.0* M_PI /(double)ntwist[i_l]),
                                   handedness,
                                   arb_twist,
                                   (void *)&(holonomies_R4[i][0][0])
                                   ); 
    };


    /* test only: "stretch" the holonomy for i=1,7 */
    /*    if( 1==i || 7==i ){
      ierr = holonomy_R3_R4( (void *)&(points_R3_SLS[i][0]), 
			     (2.0* M_PI /(((double)ntwist[i_l])*1.001)),
			     (void *)&(holonomies_R4[i][0][0]) ); 
    };
    */
    if(test || !allow_arb_twist){
      check_root_of_I( (void *)&(holonomies_R4[i][0][0]), ntwist[i_l] );
    };

  }; /*   for (i=1; i<60; i++) */


  /* identity */
  for (i=0; i< 4; i++){
      for (j=0; j< 4; j++){
	if (i==j){
	  holonomies_R4[0][i][j] = 1.0;
	}
	else{
	  holonomies_R4[0][i][j] = 0.0;
	};
      };
  };

  /* negatives */
  for (ihol=order[i_space]/2; ihol<order[i_space]; ihol++){   
    for (i=0; i< 4; i++){
      for (j=0; j< 4; j++){
	holonomies_R4[ihol][i][j] = 
	  -holonomies_R4[ihol-order[i_space]/2][i][j] ;
      };
    };

    i_l = layer(layer_index, ihol); /* find which layer this holonomy is in */
    /* find which layer this holonomy's negative is in */
    i_l_neg = layer(layer_index, ihol-order[i_space]/2);
    if(test){
      printf("\ni_holonomy= %d\n",ihol);
      printf("ihol = %d  i_l = %d i_l_neg = %d\n",ihol,i_l,i_l_neg);
      printf("ihol-order[i_space]/2 = %d \n", ihol-order[i_space]/2 );
      printf("ntwist[i_l_neg] = %d \n",ntwist[i_l_neg]);
    };
    
    if(i_space == 0 && i_l_neg == 4){
      check_root_of_I( (void *)&(holonomies_R4[ihol][0][0]), ntwist[i_l_neg] );
    };
  }; /*   for (ihol=60; ihol<120; ihol++) */

  return (0);
};


/* check that a holonomy 4x4 matrix is the n_order-th root of the
   identity, see Roukema 2005, arXiv:astro-ph/0409694, A&A 439, 479 */
int check_root_of_I( double holonomy_R4[4][4], 
		     int n_order ){

  const double tol_test2_sq = 1e-10 ;
  int i_product; /* iterate over products */
  double h_i[4][4]; /* for storing h^i */
  double h_i_plus1[4][4]; /* for storing h^(i+1) */

  int i,j;
  double xx; /* temporary work */

  int verbose = 0;

  /* initialise the product  h_i */
  for (i=0; i < 4; i++){
    for (j=0; j < 4; j++){
      h_i[i][j] = (i == j) ? 1.0 : 0.0;
    };
  };
    
  /* successively evaluate each product  h^i */
  for (i_product=1; i_product < n_order+1; i_product++){
    /* multiply h_i by the holonomy from the left */
    cblas_dgemm(101, /* CblasRowMajor */
		111, 111, /* no transpose */
		4,4,4, 1.0, (double *)holonomy_R4, 4, (double *)h_i, 4,
		0.0, (double *)h_i_plus1, 4);

    /* printf("products: h^( %d ) = \n",i_product);
       print_matrix_R4("h_i_plus1",h_i_plus1); */
    
    /* copy h_i_plus1 to h_i */
    for (i=0; i < 4; i++){
      for (j=0; j < 4; j++){
	h_i[i][j] = h_i_plus1[i][j];
      };
    };
    
    xx = 0.0 ; /* sum of elementwise square differences from identity */
    for (i=0; i < 4; i++){
      for (j=0; j < 4; j++){
	if (i == j){
	  xx += (h_i[i][j] - 1.0)* (h_i[i][j] - 1.0);
	}
	else{
	  xx += (h_i[i][j])* (h_i[i][j]);
	};
      };
    };
    xx /= 16.0;
    /* h^n_order  should give the identity */
    if ( n_order == i_product){
      if (xx > tol_test2_sq){
	printf("WARNING: h^( %d ) not identity, xx = %g \n",n_order,xx);
	print_matrix_R4("h_i",h_i);
      }
      else if(verbose){
	printf("h^( %d ) = Identity \n",n_order);  
      };
    }
    /* h^n  where n < n_order should *not* give the identity  */
    else if (i_product < n_order && xx < tol_test2_sq) {
      printf("WARNING: h^( %d ) = Identity, but %d < %d\n",
             i_product, i_product, n_order);
    }; /*       if (n_order == i_product) */

  }; /*     for (i=1; i < n_order+1; i++) */
  return (0);
};




/* Given a face centre of the fundamental domain in S^3, point_R3_SLS,
   and the corresponding perpendicular vectors point_R3_SLS_p, and the
   length of the Clifford transformation theta_holonomy, calculate the
   4x4 matrix in R^4 .

   Alternatively, use Appendix B.5 of arXiv:gr-qc/0106033v1 and the
   relation between rotational symmetries of the icosahedron embedded
   in R^3 and the PDS = S^3/I*.  

   The 0-th coordinate (0106033v1 = x-yzw convention) has been switched to the
   3-rd coordinate (xyz-w convention).

*/
int holonomy_R3_R4( /* INPUTS */
		   double point_R3_SLS[3],  /* x,y,z where R_SLS = 1 */
		   /*		   double point_R3_SLS_p[3,2]; */ /* perpendiculars */
		   
		   double theta_holonomy,   /* theta of Clifford
					       transformation in SO(4)
					       = half of theta for
					       corresponding group in
					       SO(3) in radians */

                   int    handedness,     
                   /* introduced in astromisc-0.1.18.4, needs to be tested
                      -1 = lefthanded, +1 = righthanded 
                    */

		   /* OUTPUTS */
		   double holonomy_R4[4][4] /* matrix for left
					       multiplication */
		   ){

  double a,b,c,d ; /* as per Appendix B. arXiv:gr-qc/0106033v1 */
  double sinth,xx; /* temporary work */

  int test1 = 0;
  int i;

  int i_hand; /* temporary work for handedness */
  
  a = cos(theta_holonomy);
  sinth = sin(theta_holonomy);
  b = sinth * point_R3_SLS[0];
  c = sinth * point_R3_SLS[1];
  d = sinth * point_R3_SLS[2];

  if(handedness == +1){
    i_hand = 1;
  } else if(handedness == -1){
    i_hand = 0;
  } else {
    printf("holonomy_R3_R4: Warning: invalid handedness.");
    printf("Will assume right-handed.\n");
    i_hand = 1;
  };
  

  /* TEST ONLY - THE FOLLOWING HELPS TO CHECK THE SIGN. BUT IT IS
     WRONG */
  /* 
     a=fabs(a);  
     b=fabs(b); 
     c=fabs(c); 
     d=fabs(d);  
  */

  for (i=0; i<4; i++){
    holonomy_R4[i][i] = a;
  };
  for (i=0; i<4; i++){
    xx = ( i == 3 || i == (int)fmod(2.01 + i_hand ,3) ) ? -1.0 * c : c ;
    holonomy_R4[i][(int)fmod(2.01+i+40,4)] = xx ;
  };
  for (i=0; i<4; i++){
    xx = ( i == 3 || i == (int)fmod(1.01 + i_hand ,3) ) ? -1.0 * b : b ;
    holonomy_R4[i][(int)fmod(3.01-i+40,4)] = xx ;
  };
  for (i=0; i<4; i++){
    xx = ( i == 3 || i == (int)fmod(0.01 + i_hand ,3) ) ? -1.0 * d : d ;
    holonomy_R4[i][(int)fmod(1.01-i+40,4)] = xx ;
  };

  if(test1){
    print_matrix_R4("holonomy_R4",holonomy_R4);
  }; /*   if(test) */

  return (0);
}; /* int holonomy_R3_R4  */



int space_order(int i_space){
  return order[i_space];
};

int n_spaces(void){
  return N_SPACES;
};

const char *space_name(int i_space){
  if(0 <= i_space && i_space < N_SPACES){
    return spaces[i_space];
  }else{
    return "i_space out of bounds";
  };
};


/* twist angle (first layer) is M_PI/space_ntwist if Clifford translation */
int space_ntwist(int i_space){
  int ntwist;
  switch( i_space )
    {
    case 0: 
      ntwist = octa_ntwist[1]/2; /* not 0 */
      break;
    case 1:
      ntwist = tcube_ntwist[1]/2; /* not 0 */
      break;
    case 2:
      ntwist = dodec_ntwist[1]/2; /* not 0 */
      break;
    default:
      ntwist = 1;
      printf("WARNING: i_space = %d not implemented.\n",
	     i_space);
      break;
    };
  return ntwist;
};



/* Given a face centre of the fundamental domain in S^3, point_R3_SLS,
   and the corresponding perpendicular vectors point_R3_SLS_p, and the
   length of the Clifford transformation theta_holonomy, and a twist
   angle for "circle matching", calculate the 4x4 isometry matrix in
   R^4 corresponding mathematically to a holonomy, a cancelling of the
   twist in the Clifford transformation "screw motion", and an
   insertion of the input twist angle in radians *arb_twist*.

   See the Appendices of Gausmann et al. 2001 arXiv:gr-qc/0106033v1,
   the relation between rotational symmetries of the icosahedron
   embedded in R^3 and the PDS = S^3/I*, especially Proposition A2 and
   formulae (B.1), (B.2), (B.3).

   The 0-th coordinate (0106033v1 = x-yzw convention) has been switched to the
   3-rd coordinate (xyz-w convention).

*/
int isometry_R3_R4_twist( /* INPUTS */
		   double point_R3_SLS[3],  /* x,y,z where R_SLS = 1 */
		   /*		   double point_R3_SLS_p[3,2]; */ /* perpendiculars */
		   double theta_holonomy,   /* theta of Clifford
					       transformation in SO(4)
					       = half of theta for
					       corresponding group in
					       SO(3) in radians */
                   int    handedness,     
                   /* introduced in astromisc-0.1.18.4, needs to be tested
                      -1 = lefthanded, +1 = righthanded 
                    */
                   double  arb_twist,
                   /* input arbitrary twist angle in radians; this is
                      with respect to a zero twist, i.e. the pi/5
                      twist of a first layer Poincare space holonomy
                      is handled internally in this routine and should
                      be ignored for defining the input parameter
                      *arb_twist* */

		   /* OUTPUTS */
		   double isometry_R4[4][4] /* matrix for left
					       multiplication */
		   ){

  double holonomy[4][4]; /* a true holonomy; internal use only! */
  double leftconjugator[4][4]; /* quaternion acting on left */
  double rightconjugator[4][4]; /* quaternion acting on right */
  double rightleft[4][4]; /* temporary: product right * left */
  
  double Delta_theta;

  int debug=1;
  int  ierr;

  /* angle theta (Proposition A2 Gausmann et al) to cancel the holonomy
     and insert a twist */
  Delta_theta = 0.5* (arb_twist - theta_holonomy);
  if(debug)printf("isometry_R3_R4_twist: theta_holonomy, arb_twist, Delta_theta = %g %g %g\n",
                  theta_holonomy, arb_twist, Delta_theta );
  ierr = holonomy_R3_R4( point_R3_SLS,
                         theta_holonomy,
                         handedness,
                         holonomy );
  if(debug)printf("isometry_R3_R4_twist: hol ierr = %d\n",ierr);
  
  ierr = holonomy_R3_R4( point_R3_SLS,
                         Delta_theta,
                         handedness,
                         leftconjugator );
  if(debug)printf("isometry_R3_R4_twist: hol ierr = %d\n",ierr);
  
  ierr = holonomy_R3_R4( point_R3_SLS,
                         (-Delta_theta),
                         (-handedness),
                         rightconjugator );
  if(debug)printf("isometry_R3_R4_twist: hol ierr = %d\n",ierr);
  
  /* multiply the three matrices together */
  cblas_dgemm(101, /* CblasRowMajor */
              111, 111, /* no transpose */
              4,4,4, 1.0, 
              (double *)rightconjugator, 4,
              (double *)leftconjugator, 4, 
              0.0, 
              (double *)rightleft, 4);
  cblas_dgemm(101, /* CblasRowMajor */
              111, 111, /* no transpose */
              4,4,4, 1.0, 
              (double *)rightleft, 4,
              (double *)holonomy, 4, 
              0.0, 
              (double *)isometry_R4, 4);

  return 0;
};


/** test routine **/

int test_rotate_dodec_new2old(void){
  int i;
  double a_norm;
  double points_R3_SLS[3]; /* a point in R3 on the SLS */
  double carto1,carto2,carto3; /* rotated R3 vector */
  double gl,gb;  /* face centre */
  
  i=1;
  a_norm = sqrt( dodec_symm[i][0] * dodec_symm[i][0] + 
                 dodec_symm[i][1] * dodec_symm[i][1] + 
                 dodec_symm[i][2] * dodec_symm[i][2] );
  
  for(i=1;i<13;i++){
    printf(" i = %d : ",i);
    points_R3_SLS[0] = dodec_symm[i][0] / a_norm ;
    points_R3_SLS[1] = dodec_symm[i][1] / a_norm ;
    points_R3_SLS[2] = dodec_symm[i][2] / a_norm ;


    rotate_dodec_new2old(
                         /* 0.0,90.0,  0.0, */
                         184.0, 62.0, 34.0,  /* RBSG08 = arXiv:0801.0006 */
                         points_R3_SLS[0],
                         points_R3_SLS[1],
                         points_R3_SLS[2],
                         &carto1,&carto2,&carto3);
    printf("test: %5.2f  %5.2f  %5.2f    ",carto1,carto2,carto3);
    printf(" %8.3f \n",carto1*carto1 + carto2*carto2 + carto3*carto3);

    /* rough xyz_rda, TODO: should do properly in C with tolerances,
       but here this is not testing tolerances
     */
    gb = asin(carto3) /M_PI_180;
    gl = acos( carto1 / sqrt( 1.0001 - carto3*carto3 )) /M_PI_180;
    if(carto2 < 0) gl = 360.0 - gl;
    printf("face: %6.1f %6.1f \n",gl,gb);

  };
  return 0;
}






/** test routine **/
#define NA 10
#define NB 3
int test_NP_rotate_R3(void){

  /*  definition of rotation in DEGREES */
  double gal_l[NA]= {0,0,0,0,     45,45, 90, 90, 180, 270};  
  double gal_b[NA]= {0,10,20,-20, 0,45, 0,   90, 0, 0}; 
  double cart1[NB]= {1,0,0}; /* input position in cartesian coords */
  double cart2[NB]= {0,1,0}; double cart3[NB]={0,0,1}; 
  double carto1; /* output position in cart coords */
  double carto2; double carto3;
  
  int i,j;

  for(i=0;i<NA;i++){
    printf("gal_l = %.2f gal_b = %.2f",gal_l[i],gal_b[i]);

    /*  
        printf("\n");
        NP_rotate_R3( gal_l[i], gal_b[i],
                  cos(gal_b[i] *M_PI_180) *cos(gal_l[i] *M_PI_180),
                  cos(gal_b[i] *M_PI_180) *sin(gal_l[i] *M_PI_180),
                  sin(gal_b[i] *M_PI_180),
                  &carto1,&carto2,&carto3 );
    printf("  gl, gb sent to:  %.2f %.2f %.2f\n\n",carto1,carto2,carto3);
    */

    printf(" = [%.2f %.2f %.2f] \n",
           cos(gal_b[i] *M_PI_180) *cos(gal_l[i] *M_PI_180),
           cos(gal_b[i] *M_PI_180) *sin(gal_l[i] *M_PI_180),
           sin(gal_b[i] *M_PI_180));
    NP_rotate_R3( gal_l[i], gal_b[i],
                  0.0, 0.0, 1.0,
                  &carto1,&carto2,&carto3 );
    printf("  NP sent to:  %.2f %.2f %.2f\n\n",carto1,carto2,carto3);


    for(j=0;j<NB;j++){
      printf("  input: cart123 = %.2f %.2f %.2f\n",cart1[j],cart2[j],cart3[j]);
      NP_rotate_R3( gal_l[i], gal_b[i],
                    cart1[j],cart2[j],cart3[j],
                    &carto1,&carto2,&carto3 );
      printf("  output: carto123 = %.2f %.2f %.2f\n\n",carto1,carto2,carto3);
    };
  };


  /*  gsl_rng_env_setup();
  T= gsl_rng_default;
  r= gsl_rng_alloc(T);
  */

  /*   x= gsl_rng_uniform(r); */
  return 0;
}


/* flat case (at least, T^3) */

  
/* returns number of holonomies available */
int genholon_flat(int i_space,                            /* which space, unused(?) */ 
		  double holonomies_R4[MAX_ORDER][4][4],
		  struct holon_coords_s holon_coords
		  ){
  int n_holonomies; /* return value: how many holonomies are available */
  int i_axis, i,j,k, i_hol;
  double factor_ith_axis; /* i-1, j-1, or k-1  matched to axis number i_axis */

  int debug=0;

  n_holonomies = 27;
  if(3!=holon_coords.n_axes){
    printf("genholon_flat Warning: holon_coords.n_axes = %d\n",
	   holon_coords.n_axes);
  };
  for(i_hol=0; i_hol < n_holonomies; i_hol++){
    /* default: identity matrix */
    for(i=0;i<4;i++){
      for(j=0;j<4;j++){
	if(i==j){
	  holonomies_R4[i_hol][i][j] = 1.0;
	}else{
	  holonomies_R4[i_hol][i][j] = 0.0;
	};
      };
    };
  };

  /* vector holonomies */
  /* 3x3 cube = 27 holonomies; identity + 3 already done */
  for(i=0; i<3; i++){
    for(j=0; j<3; j++){
      for(k=0; k<3; k++){
	/* check that the numbering system is consistent */
	if(27==n_holonomies){
	  i_hol = 9*k + 3*j + i;
	}else{
	  printf("genholon_flat error: n_holonomies = %d not programmed.\n",
		 n_holonomies);
	  exit(EXIT_FAILURE);
	};
	
	/* add a vector for each direction */
	if(debug){ printf("AA. i = %d  j= %d k=%d \n",i,j,k); };
	for(i_axis=0;i_axis<3;i_axis++){

	  /* TODO when online: check syntax of case and/or switch statement! */
	  if(i_axis==0){
	    factor_ith_axis = i-1.0;
	  }else if(i_axis==1){
	    factor_ith_axis = j-1.0;
	  }else if(i_axis==2){
	    factor_ith_axis = k-1.0;
	  };
	  
	  if(debug){
	    printf("A. i_hol = %d, i_axis= %d\n",i_hol,i_axis);
	    print_matrix_R4("holonomies_R4[i_hol]",holonomies_R4[i_hol]);
	    printf("\n");
	  };

	  holonomies_R4[ i_hol ][0][3] += 
	    factor_ith_axis *
	    holon_coords.L_axis[ i_axis ] *
	    cos( (holon_coords.gal_l[ i_axis ])*M_PI/180.0 ) *
	    cos( (holon_coords.gal_b[ i_axis ])*M_PI/180.0 ) ; /* x shift */
	  if(debug){
	    printf("B. i_hol = %d, i_axis= %d\n",i_hol,i_axis);
	    print_matrix_R4("holonomies_R4[i_hol]",holonomies_R4[i_hol]);
	    printf("\n");
	  };

	  holonomies_R4[ i_hol ][1][3] += 
	    factor_ith_axis *
	    holon_coords.L_axis[ i_axis ] *
	    sin( (holon_coords.gal_l[ i_axis ])*M_PI/180.0 ) *
	    cos( (holon_coords.gal_b[ i_axis ])*M_PI/180.0 ) ; /* y shift */

	  if(debug){
	    printf("C. i_hol = %d, i_axis= %d\n",i_hol,i_axis);
	    print_matrix_R4("holonomies_R4[i_hol]",holonomies_R4[i_hol]);
	    printf("\n");
	  };

	  holonomies_R4[ i_hol ][2][3] += 
	    factor_ith_axis *
	    holon_coords.L_axis[ i_axis ] *
	    sin( (holon_coords.gal_b[ i_axis ])*M_PI/180.0 ) ; /* z shift */

	  if(debug){
	    printf("D. i_hol = %d, i_axis= %d\n",i_hol,i_axis);
	    print_matrix_R4("holonomies_R4[i_hol]",holonomies_R4[i_hol]);
	    printf("\n");
	  };

	}; /* for(i_axis=0;i_axis<3;i_axis++) */
      }; /* for(k=0; k<3; k++) */
    }; /*    for(j=0; j<3; j++) */
  }; /*  for(i=0; i<3; i++) */
  
  /* temporary hack: copy i_hol 0 to 13, and make 0 the identity, so
     that the identity has i_hol=0 */
  holonomies_R4[ 13 ][0][3] =   holonomies_R4[ 0 ][0][3] ; 
  holonomies_R4[ 13 ][1][3] =   holonomies_R4[ 0 ][1][3] ; 
  holonomies_R4[ 13 ][2][3] =   holonomies_R4[ 0 ][2][3] ; 
  holonomies_R4[ 0 ][0][3] =  0.0;
  holonomies_R4[ 0 ][1][3] =  0.0;
  holonomies_R4[ 0 ][2][3] =  0.0;

  return n_holonomies;
}

/* this is almost certainly redundant, written elsewhere */
double d_R3(double x1, double y1, double z1, double x2, double y2, double z2){
  double xx,yy,zz;
  double distance; /* return value */
  xx=x1-x2;
  yy=y1-y2;
  zz=z1-z2;
  distance = sqrt(xx*xx + yy*yy + zz*zz);
  return distance;
};



int test_genholon_flat(void){
  int n_holonomies; /*  how many holonomies are available */

  /* inputs to genholon_flat */
  int i_space =-99;
  double holonomies_R4[MAX_ORDER][4][4];
  struct holon_coords_s holon_coords;

  /* local */
  int i_axis; /* for the three axes */
  int i_hol;  /* for the holonomies */

  /* Aurich 2008 CQG, 25, 225017, arXiv:0803.2130,  */
  double T3_gal_l[6] = {6.0, 17.0, 107.0, 186.0, 197.0, 287.0};
  double T3_gal_b[6] = {77.0, -13.0, 3.0,  -77.0, 13.0, -3.0};
  double T3_L = 10.0; /* 11512.0; */

  double vec_in[3][4] = { {17.0, 0.0, 0.0, 1.0}, 
			  {0.0, 17.0, 0.0, 1.0}, 
			  {0.0, 0.0, 17.0, 1.0} }; 
  int i_input;
  double vec_out[4];
  int j; /* trivial local only */

  holon_coords.n_axes=3;
  for(i_axis=0; i_axis<holon_coords.n_axes; i_axis++){
    holon_coords.gal_l[i_axis] = T3_gal_l[i_axis];
    holon_coords.gal_b[i_axis] = T3_gal_b[i_axis];
    holon_coords.L_axis[i_axis] = T3_L;
  };


  n_holonomies = genholon_flat(i_space, holonomies_R4, holon_coords);
  printf("Called genholon_flat.  n_holonomies=%d\n",n_holonomies);

  for(i_input=0;i_input<3;i_input++){
    for(i_hol=0;i_hol<n_holonomies; i_hol++){
      printf("i_hol = %d\n",i_hol);
      print_matrix_R4("holonomies_R4[i_hol]",holonomies_R4[i_hol]);
      printf("\n");
      for(j=0;j<4;j++){ vec_out[j]=0.0; };

      /* multiply vec_in[...]  by the  i_hol-th holonomy from the left */
      cblas_dgemv(CblasRowMajor, /* 101, CblasRowMajor */
		  CblasNoTrans, /* 111,  no transpose  CblasNoTrans */
		  4,4, 1.0, (double *)holonomies_R4[i_hol], 4, 
		  (double *)vec_in[i_input], 1,
		  0.0, (double *)vec_out, 1);

      printf("i_hol = %d, vec_in = ", i_hol);
      for(j=0;j<4;j++){ printf("%f ",vec_in[i_input][j]); };
      printf(",\n          vec_out = ");
      for(j=0;j<4;j++){ printf("%f ",vec_out[j]); };
      printf("\n");

      printf("d= %g\n\n", d_R3(vec_in[i_input][0], vec_in[i_input][1], vec_in[i_input][2],
			   vec_out[0], vec_out[1], vec_out[2]) );
    };
  };
  

  return 0;
};

