/* 
   astromisc - library with miscellaneous GPL tools for astronomy research
   genholon.h - header file for defining holonomies of the Poincare space

   Copyright (C) 2004 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

/* don't define these multiple times */

/* #ifndef ASTROMISC_HOLONOMIES
   #define ASTROMISC_HOLONOMIES 1 
*/
/* #endif */
/* endif: #ifndef ASTROMISC_HOLONOMIES */


#ifndef GENHOLON_SPHERICAL_WELL_PROPORTIONED
#define GENHOLON_SPHERICAL_WELL_PROPORTIONED 1

/* how many different spherical spaces are included so far */
#define N_SPACES 3 


char const *spaces[N_SPACES] = {
  "octa",     /* 0 = octahedral space */
  "tcube",     /* 1 = truncated cube space */
  "dodec"     /* 2 = Poincare dodecahedral space */
} ;

/* maximum order of a space, and order of individual spaces */
#ifndef MAX_ORDER
#define MAX_ORDER 120 
#endif

/* maximum number of axes for defining a holonomy */
#ifndef MAX_AXES
#define MAX_AXES 3
#endif



int const order[N_SPACES] = {
  24,     /* 0 = octahedral space */
  48,     /* 1 = truncated cube space */
  120     /* 2 = Poincare dodecahedral space */
} ;
  

/* maximum number of "layers", including identity */
/* (In fact, the equatorial layer is divided into e.g. two, e.g. for the PDS,
   there are really only 9 layers, not 10.) */
#define MAX_LAYER 10 
int const n_layers[N_SPACES] = {
  6,     /* 0 = octahedral space */
  10,    /* 1 = truncated cube space */
  10     /* 2 = Poincare dodecahedral space */
} ;




double const octa_symm[12][3] = {
  /* ZEROTH layer i= 0 = identity; do not use this */
  { 0.0, 0.0, 0.0 }  ,
  /* FIRST layer (8): i = 1...8 */
  { 1.0,  1.0, 1.0 },
  { 1.0,  -1.0, 1.0 },
  { 1.0,  1.0, -1.0 },
  { 1.0,  -1.0, -1.0 },
  /* negatives */
  { -1.0,  1.0, 1.0 },
  { -1.0,  -1.0, 1.0 },
  { -1.0,  1.0, -1.0 },
  { -1.0,  -1.0, -1.0 },
  /* SECOND (middle) layer (3): i=9...11 */
  { 1.0, 0.0, 0.0 },
  { 0.0, 1.0, 0.0 },
  { 0.0, 0.0, 1.0 }
};


/* indexing the different layers */
/* * the middle layer is written here as two "layers" */
int const octa_layer_index[6][2] = {
  {0, 0},
   {1, 8},
    {9, 11}, /* half of the middle layer */
    {21, 23}, /* the other half of the middle layer */
   {13, 20},
  {12, 12}
};
  
int const octa_ntwist[3] = { 
  1, 
   6, 
    4
};






double const tcube_symm[24][3] = {
  /* ZEROTH layer i= 0 = identity; do not use this */
  { 0.0, 0.0, 0.0 }  ,
  /* FIRST layer (6): i = 1...6 */
  { 1.0, 0.0, 0.0 },
  { 0.0, 1.0, 0.0 },
  { 0.0, 0.0, 1.0 },
  /* negatives */
  { -1.0, 0.0, 0.0 },
  { 0.0, -1.0, 0.0 },
  { 0.0, 0.0, -1.0 },
  /* SECOND layer (8): i=7...14 */
  { 1.0,  1.0, 1.0 },
  { 1.0,  -1.0, 1.0 },
  { 1.0,  1.0, -1.0 },
  { 1.0,  -1.0, -1.0 },
  /* negatives */
  { -1.0,  1.0, 1.0 },
  { -1.0,  -1.0, 1.0 },
  { -1.0,  1.0, -1.0 },
  { -1.0,  -1.0, -1.0 },
  /* THIRD (middle-A) layer (6): i = 15...20 */
  { 1.0, 1.0, 0.0 },
  { 0.0, 1.0, 1.0 },
  { 1.0, 0.0, 1.0 },
  /* negatives */
  { -1.0, 1.0, 0.0 },
  { 0.0, -1.0, 1.0 },
  { 1.0, 0.0, -1.0 },
  /* SECOND (middle) layer (3): i=21...23 */
  { 1.0, 0.0, 0.0 },
  { 0.0, 1.0, 0.0 },
  { 0.0, 0.0, 1.0 }
};


/* indexing the different layers */
/* * the middle layer is written here as two "layers" */
int const tcube_layer_index[10][2] = {
  {0, 0},
   {1, 6},
    {7, 14},
     {15, 20}, /* half of the A middle layer */
     {21, 23}, /* half of the B middle layer */
     {45, 47}, /* other half of the B middle layer */
     {39, 44}, /* the other half of the A middle layer */
    {31, 38},
   {25, 30},
  {24,24}
};
  
int const tcube_ntwist[5] = { 
  1, 
   8, 
    6,
     4,
      4
};




#define PHI_GOLDEN 0.61803398874989490253
#define PHI_GOLDEN_1 1.61803398874989490253
#define PHI_GOLDEN_2 2.61803398874989490253

#ifndef HOL_SYMMETRIES
#define HOL_SYMMETRIES 1
double const dodec_symm[60][3] = {
  /* ZEROTH layer i= 0 = identity; do not use this */
  { 0.0, 0.0, 0.0 }  ,
  /* FIRST layer (12): i = 1...12 */
  { PHI_GOLDEN,  1.0, 0.0 },
  { PHI_GOLDEN, -1.0, 0.0 },
  { 0.0, PHI_GOLDEN,  1.0 },
  { 0.0, PHI_GOLDEN, -1.0 },
  {  1.0, 0.0, PHI_GOLDEN },
  { -1.0, 0.0, PHI_GOLDEN },
  /* negatives */
  { -PHI_GOLDEN, -1.0, 0.0 },
  { -PHI_GOLDEN,  1.0, 0.0 },
  { 0.0, -PHI_GOLDEN, -1.0 },
  { 0.0, -PHI_GOLDEN,  1.0 },
  { -1.0, 0.0, -PHI_GOLDEN },
  {  1.0, 0.0, -PHI_GOLDEN },
  /* SECOND layer (20): i=13...32 */
  { 1.0,  1.0,  1.0 },
  { 1.0, -1.0,  1.0 },
  { 1.0,  1.0, -1.0 },
  { 1.0, -1.0, -1.0 },
  { 0, PHI_GOLDEN_2,  1.0 },
  { 0, PHI_GOLDEN_2, -1.0 },
  {  1.0, 0, PHI_GOLDEN_2 },
  { -1.0, 0, PHI_GOLDEN_2 },
  { PHI_GOLDEN_2,  1.0, 0 },
  { PHI_GOLDEN_2, -1.0, 0 },
  /* negatives */
  { -1.0, -1.0, -1.0 },
  { -1.0,  1.0, -1.0 },
  { -1.0, -1.0,  1.0 },
  { -1.0,  1.0,  1.0 },
  { 0, -PHI_GOLDEN_2, -1.0 },
  { 0, -PHI_GOLDEN_2,  1.0 },
  { -1.0, 0, -PHI_GOLDEN_2 },
  {  1.0, 0, -PHI_GOLDEN_2 },
  { -PHI_GOLDEN_2, -1.0, 0 },
  { -PHI_GOLDEN_2,  1.0, 0 },
  /* THIRD layer (12): i=33...44 (identical to first layer) */
  { PHI_GOLDEN,  1.0, 0.0 },
  { PHI_GOLDEN, -1.0, 0.0 },
  { 0.0, PHI_GOLDEN,  1.0 },
  { 0.0, PHI_GOLDEN, -1.0 },
  {  1.0, 0.0, PHI_GOLDEN },
  { -1.0, 0.0, PHI_GOLDEN },
  /* negatives */
  { -PHI_GOLDEN, -1.0, 0.0 },
  { -PHI_GOLDEN,  1.0, 0.0 },
  { 0.0, -PHI_GOLDEN, -1.0 },
  { 0.0, -PHI_GOLDEN,  1.0 },
  { -1.0, 0.0, -PHI_GOLDEN },
  {  1.0, 0.0, -PHI_GOLDEN },
  /* FOURTH layer (15) i=45...59 */
  { 1.0, 0.0, 0.0 },
  { 0.0, 1.0, 0.0 },
  { 0.0, 0.0, 1.0 },
  {  PHI_GOLDEN,  PHI_GOLDEN_1,  1.0 },
  {  PHI_GOLDEN, -PHI_GOLDEN_1,  1.0 },
  {  1.0,  PHI_GOLDEN,  PHI_GOLDEN_1 },
  {  1.0,  PHI_GOLDEN, -PHI_GOLDEN_1 },
  {  PHI_GOLDEN_1,  1.0,  PHI_GOLDEN },
  { -PHI_GOLDEN_1,  1.0,  PHI_GOLDEN },
  /* negatives */
  { -PHI_GOLDEN, -PHI_GOLDEN_1,  1.0 },
  { -PHI_GOLDEN,  PHI_GOLDEN_1,  1.0 },
  {  1.0, -PHI_GOLDEN, -PHI_GOLDEN_1 },
  {  1.0, -PHI_GOLDEN,  PHI_GOLDEN_1 },
  { -PHI_GOLDEN_1,  1.0, -PHI_GOLDEN },
  {  PHI_GOLDEN_1,  1.0, -PHI_GOLDEN }
};


/* indexing the different layers */
/* * the middle layer is written here as two "layers" */
int const dodec_layer_index[10][2] = {
  {0, 0},
   {1, 12},
    {13, 32},
     {33, 44},
      {45, 59},   /* half of the middle layer */
      {105, 119}, /* the other half of the middle layer */
     {93, 104},
    {73, 92},
   {61, 72},
  {60, 60}
};
  
int const dodec_ntwist[5] = { 
  1, 
   10, 
     6, 
      5,
       4
};


#endif



/* which layer is PDS (oholonomy i in ? */
int layer(int layer_index[MAX_LAYER][2],  /* INPUT, do not modify */
	  int i /* input, holonomy number */
	  );



/* check that a holonomy 4x4 matrix is the n_order-th root of the
   identity, see Roukema 2005, arXiv:astro-ph/0409694, A&A 439, 479 */
int check_root_of_I( double holonomy_R4[4][4], 
		     int n_order );


int holonomy_R3_R4( /* INPUTS */
		   double point_R3_SLS[3],  /* x,y,z where R_SLS = 1 */
		   double theta_holonomy,   /* theta of Clifford
					       transf */
                   int    handedness,      
		   /* OUTPUTS */
		   double holonomy_R4[4][4] /* matrix for left
					      multiplication */
		   );

int isometry_R3_R4_twist( /* INPUTS */
		   double point_R3_SLS[3],  /* x,y,z where R_SLS = 1 */
		   double theta_holonomy,   /* theta of Clifford
					       transf */
                   int    handedness,      
                   double arb_twist,        /* arbitrary twist angle: not holonomy */
		   /* OUTPUTS */
		   double holonomy_R4[4][4] /* matrix for left
					      multiplication */
		   );


/* calculate the pairwise products of holonomies as 4x4 matrices and
   find out which element of the group results */
int product_table( int i_space,                    /* which space ? */
		   double holonomies_R4[MAX_ORDER][4][4], 
		   int holonomy_products_i[MAX_ORDER][MAX_ORDER] );


/* calculate the holonomies as 4x4 matrices generated by a fixed
   pair of would-be generators, find out which element of the group 
   results in each case, in order to see if the full group is
   generated */
int product_table_generator_pair(int i_space,                    /* which space ? */ 
                                 
                                 double holonomies_R4[MAX_ORDER][4][4], 
                                 int generators[2],
                                 int max_gen_order
                                 );


#endif
/* #ifndef GENHOLON_SPHERICAL_WELL_PROPORTIONED */

/* This should PRIMARILY be defined in holonomies.h. See the TODO below. */
#ifndef GENHOLON_FN
#define GENHOLON_FN 1
/* calculate the MAX_ORDER holonomies as 4x4 matrices */
int genholon(int i_space,                            /* which space */ 
	     double holonomies_R4[MAX_ORDER][4][4],
             double gal_l, /* face centre if dodecahedral space */
             double gal_b,
             double theta, /* third orientation parameter if dodecspace */
             int    handedness,  /* -1 = lefthanded, +1 = righthanded */
             int    allow_arb_twist, /* 0 = don't allow, 1 = allow */
             double arb_twist  /* arbitrary twist angle */
             ); 

struct holon_coords_s{
  int n_axes;
  double gal_l[MAX_AXES]; /* gal longitude of axes */
  double gal_b[MAX_AXES]; /* gal latitude of axes */
  double L_axis[MAX_AXES]; /* comoving lengths of axes */
};


/* calculate the MAX_ORDER holonomies as 4x4 matrices */
int genholon_flat(int i_space,                            /* which space */ 
		  double holonomies_R4[MAX_ORDER][4][4],
		  struct holon_coords_s holon_coords
		  ); 

int check_genholon_flat(void);
/* TODO: try (again) to avoid doubling up these parts of the header files!!! */

#endif
/* endif: #ifndef GENHOLON_FN */

