/*
## Copyright (C) 1995, 1996, 1997  Kurt Hornik
## Copyright (C) 2009  Boud Roukema minor adaptation - octave2.9 to C 
##
## This file is adapted from a part of Octave.
##
## Octave is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2, or (at your option)
## any later version.
##
## Octave is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Octave; see the file COPYING.  If not, write to the Free
## Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
## 02110-1301, USA.


## Author: KH <Kurt.Hornik@wu-wien.ac.at>
## Description: Two-sample Kolmogorov-Smirnov test
*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gsl/gsl_sort.h>

#include "astromisc.h"

/* TODO: dynamic rather than hardwired memory allocation */
#define MAXN 1000



int kolmogorov_smirnov_test_2(double *x,    /* inputs */
			      int   n_x,
			      double *y, 
			      int   n_y,
			      char *alt,      /* inputs: TODO */
			      double *pval,   /* outputs: TODO */
			      double *ks,     /* outputs: DONE */
			      double *d_ks    /* outputs: DONE */
			      ){

  /* local use */
  double big_n ;

  double xy_sorted[2*MAXN];
  size_t xy_indices[2*MAXN];

  double count[2*MAXN];
  double cumul[2*MAXN];

  double d_xy_sorted[2*MAXN];

  int i;
  int ioffset=0;
  int iwarn=0;

  int test1=0;

  if( n_x > MAXN || n_y > MAXN){
    printf("ERROR. n_x = %d || n_y = %d bigger than MAXN = %d.\n",
	   n_x,n_y,MAXN);
    printf("Increase MAXN in ks.c or improve ks.c in another way.\n");
  };
  
  big_n  = (double)(n_x * n_y) / (double)(n_x + n_y);


  /*  [s, i] = sort ([x; y]); */
  for (i=0;i<n_x;i++){
    xy_sorted[i] = x[i];
  };
  for (i=n_x;i<n_x+n_y;i++){
    xy_sorted[i] = y[i-n_x];
  };

  /* get permutation indices, don't yet change xy_sorted */
  gsl_sort_index(xy_indices,xy_sorted,1,n_x+n_y);

  /* change xy_sorted into sorted order */
  gsl_sort(xy_sorted,1,n_x+n_y);

  /*  count (find (i <= n_x)) = 1 / n_x;
      count (find (i > n_x)) = - 1 / n_y; */

  for (i=0;i<n_x+n_y;i++){
    if(xy_indices[i] < n_x){
      count[i] = 1.0/(double)n_x;
    }else{
      count[i] = -1.0/(double)n_y;
    };
  };

  if(test1){
    printf("count_t = [\n");
    for (i=0;i<n_x+n_y;i++){
      printf("%f  ",count[i]);
    };
    printf("]\n");
  }

  /*  z = cumsum (count); */
  /* ds = diff (s); */
  cumul[0]= count[0];
  for (i=1;i<n_x+n_y;i++){
    cumul[i]= cumul[i-1]+count[i];
    d_xy_sorted[i-1]= xy_sorted[i] - xy_sorted[i-1];
  };

  if(test1){
    printf("cumul_t = [\n");
    for (i=0;i<n_x+n_y;i++){
      printf("%f  ",cumul[i]);
    };
    printf("]\n");
    printf("d_xy_sorted_t = [\n");
    for (i=0;i<n_x+n_y-1;i++){
      printf("%f  ",d_xy_sorted[i]);
    };
    printf("]\n");
  }


  /*  if (any (ds == 0))
    ## There are some ties, so keep only those changes.
    warning ("cannot compute correct p-values with ties")
    elems = [find(ds); n_x+n_y];
    z = z(elems);
  endif 
  */

  *d_ks= -9e9;
  for (i=0; i+ioffset<n_x+n_y; i++){
    if(i+ioffset<n_x+n_y-1){
      if( d_xy_sorted[i+ioffset] == 0.0 ){
	ioffset ++;
	if(ioffset == 1 && iwarn ==0){
	  printf("Warning: possible error in computing correct p-values with ties.\n");
	  printf("xy_sorted[%d] =  %f  xy_sorted[%d] =  %f  \n",
		 i+ioffset,xy_sorted[i+ioffset],
		 i+ioffset-1,xy_sorted[i+ioffset-1]
		 );
	  iwarn++;
	};
      };
      cumul[i] = fabs(cumul[i+ioffset]);
      if(test1)printf("new cumul[%d] = %f\n",i,cumul[i]);
      if(cumul[i] > *d_ks){
	*d_ks = cumul[i] ;
      };
    }else{
      if(fabs(cumul[n_x+n_y-1]) > *d_ks){
	*d_ks = cumul[i] ;
      };
    };

    if(test1){
      printf(" i = %d  *d_ks = %f \n",
	     i,*d_ks);
    };
  };

  *ks  = sqrt (big_n) * (*d_ks);
  /* TODO: pval = 1 - kolmogorov_smirnov_cdf (ks); */

  return 0;
};


/* function [pval, ks, d] = kolmogorov_smirnov_test_2 (x, y, alt) */

/* TODO: checks
  if (nargin < 2 || nargin > 3)
    usage ("[pval, ks] = kolmogorov_smirnov_test_2 (x, y, tol)");
  endif

  if (! (isvector (x) && isvector (y)))
    error ("kolmogorov_smirnov_test_2: both x and y must be vectors");
  endif

  if (nargin == 2)
    alt = "!=";
  else
    if (! ischar (alt))
      error ("kolmogorov_smirnov_test_2: alt must be a string");
    endif
  endif
*/

/*  
  if (strcmp (alt, "!=") || strcmp (alt, "<>"))
    d    = max (abs (z));
    ks   = sqrt (n) * d;
    pval = 1 - kolmogorov_smirnov_cdf (ks);
  elseif (strcmp (alt, ">"))
    d    = max (z);
    ks   = sqrt (n) * d;
    pval = exp (-2 * ks^2);
  elseif (strcmp (alt, "<"))
    d    = min (z);
    ks   = -sqrt (n) * d;
    pval = exp (-2 * ks^2);
  else
    error ("kolmogorov_smirnov_test_2: option %s not recognized", alt);
  endif

  if (nargout == 0)
    printf ("  pval: %g\n", pval);
  endif
*/




void CHKKS2( void
	     ){

  /*  double x[4] = { 3, 4, 7.0, 9.0} ;   
  int   n_x=4;
  double y[3] = { 3.2, 4.0, 7.5 } ; 
  int   n_y=3;

  double x[7] = { 3, 4, 19.1, 7.0, -3.0, 9.0, -5e3} ; 
  int   n_x=7;
  double y[5] = { 3.2, -5.0, 7.5, 4.0, 7.5 } ; 
  int   n_y=5;
  */

  double x[7] = { 3, 4, -19.1, 7.0, -3.0, 9.0, -5e3} ;
  int   n_x=7;
  double y[5] = { 3.2, -59.0, 7.5e2, -4.0, 7.5 } ; 
  int   n_y=5;


  char *alt = "<>";      /* inputs: TODO */
  double pval;   /* outputs: TODO */
  double ks;     /* outputs: DONE */
  double d_ks;    /* outputs: DONE */

  kolmogorov_smirnov_test_2(x,n_x, y,n_y, alt,
			    &pval, &ks, &d_ks);
  printf("d_ks = %f \n",d_ks);
};

