C     ylm - spherical harmonics, e.g. real-valued, using GSL
C     Copyright (C) 2008 B. Roukema <boud at astro.uni.torun.pl>
C     
C     This program is free software; you can redistribute it and/or modify
C     it under the terms of the GNU General Public License as published by
C     the Free Software Foundation; either version 2 of the License, or
C     (at your option) any later version.
C     
C     This program is distributed in the hope that it will be useful,
C     but WITHOUT ANY WARRANTY; without even the implied warranty of
C     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C     GNU General Public License for more details.
C     
C     You should have received a copy of the GNU General Public License
C     along with this program; if not, write to the Free Software
C     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
C  
C     See also http://www.gnu.org/licenses/gpl.html

      real*8 function ylm8(l_in,m_in, phi,  the_in, inorth)

      implicit real*8(a-h,o-z)

      integer    l_in,m_in      ! l, m of spherical harmonic
      real*8     phi            ! input longitude in radians
      real*8     the_in         ! input latitude radians; convention: see inorth
      integer    inorth         ! multiple of pi/2 of North pole = -1,0,1 or 2 

      parameter(pi=3.141592659,pi_2 = 0.5*pi)


     
c     ---- convert the_in to theta ----
      if(inorth.eq.1)then
         theta= pi*0.5 - the_in
      elseif(inorth.eq.-1)then
         theta= pi*0.5 + the_in
      elseif(inorth.eq.0)then
         theta= the_in
      elseif(inorth.eq.2)then
         theta = pi - the_in
      else
         print*,'ylm ERROR: inorth = ',inorth,' not recognised.'
         stop
      endif

c     --- test only ----
c      do i=1,10
c         print*,'i, fctgsl(i) = ',i
c         print*,nint(fctgsl(i))
c      enddo
c      i=69
c      print*,'i, fctgsl(i) = ',i,fctgsl(i)

c     --- check l,m ---
      if (abs(m_in).gt.l_in)then
         print*,'ylm ERROR: l,m =', l_in,m_in
         stop
      endif


c     --- call GSL plm routine ---
      p= plmgsl(l_in,abs(m_in),dble(cos(theta))) 

c     --- factor of sqrt of factorials 
c     --- TODO: fatal errors in the following can occur if fctgsl is faulty ---
      f= sqrt(
     _     fctgsl(l_in - m_in) / fctgsl(l_in + m_in)
     _     * (2.0*l_in + 1.0) * 0.25 / pi
     _     )

      if ( 2*(m_in/2).ne.m_in ) f = -f  ! multiply by (-1)^m_in

      ylm8 = f * p * cos(m_in * phi)

      return
      end


      subroutine chkylm()
      implicit real*8(a-h,o-z)

      parameter(pi=3.141592659,pi_2 = 0.5*pi)
      parameter(pi_180=pi/180.0)

      real*8     ylm8
      external   ylm8

c      read(5,*)l,m,phi,theta
c      x = ylm8(l,m, phi, theta, 1)

c      x = ylm8(3,2, 1.0d0, 1.0d0, 1)
c      print*,'x= ',x
c      stop

      glong1=0.0
c      glong2=360.0
      glat1=-90.0
c      glat2=+90.0
      maxn= 50
      step= 180.0/real(maxn)

      read(5,*)l,m

      do ib=1,maxn
         glat= glat1 + (real(ib)-0.5)*step
         do il=1,maxn*2
            glong= glong1 + (real(il)-0.5)*step
            y = ylm8(l,m, glong*pi_180, glat*pi_180, 1)
c     print*,glong,glat,y
            write(6,'(f13.5,a,f13.5,a,g14.5,$)')glong,' ',glat,' ',y
            print*,''
         enddo
c         print*,''
      enddo

c     --- pipe this through proj, e.g.
c     | proj +proj=laea +lat_0=-90 +lon_0=0 -m 7e-8 
c     ---

      return
      end
